/**
 *  Copyright (c) 2010 A.G.Docherty. All Rights Reserved.
 */

package uk.co.keang.sudoku;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.EventQueue;
import java.awt.GridLayout;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.List;

import javax.swing.DefaultListModel;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSlider;
import javax.swing.JToggleButton;
import javax.swing.ListSelectionModel;
import javax.swing.border.EmptyBorder;
import javax.swing.border.TitledBorder;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

import uk.co.keang.sudoku.Puzzle.Cell;
import uk.co.keang.sudoku.Puzzle.PuzzleDifficulty;
import uk.co.keang.sudoku.event.IllegalGridException;
import uk.co.keang.sudoku.event.SolutionEvent;
import uk.co.keang.sudoku.event.SolutionEvent.DisplayOrder;

/**
 * @author A.G.Docherty
 *
 */
class SolutionPanel extends JPanel implements GridChangeListener
{
private static final long serialVersionUID = -7068345439151936895L;
private int frameDelay = 500;
private final SudokuSolver sudokuSolver;
private final SequenceBuffer seqBuf;
private final SudokuGrid gridGUI;
private final Player player;
private JList eventList;
private JButton playBtn;
private JToggleButton pauseBtn;
private JButton stopBtn;
public boolean enableListSelectionEvents;
public List<SolutionEvent> coalescedEvents;
private transient ListSelectionListener eventListlistener;
private JButton stepBtn;
private boolean autoSolve;
private JLabel statusLbl;

public SolutionPanel(SudokuSolver puzzle, SequenceBuffer sequenceBuffer, SudokuGrid gridPanel)
	{
	sudokuSolver = puzzle;
	seqBuf = sequenceBuffer;
	gridGUI = gridPanel;

	gridGUI.addGridChangeListener(this);
	
	initPanel();

	// compute the result
	try
	    {
	    sudokuSolver.compute();
	    }
    catch ( IllegalGridException exc )
	    {
	    JOptionPane.showMessageDialog(this, "The sudoku puzzle is illegal.\n"+exc.getMessage(), "Illegal Puzzle", JOptionPane.ERROR_MESSAGE);
	    }
    
	coalescedEvents = seqBuf.getCoalescedEvents();

	player = new Player();
	
	// init the grid to the start position
	final Puzzle puzz = seqBuf.getStartGrid();

	EventQueue.invokeLater(new Runnable()
		{
		@Override
		public void run()
			{
			gridGUI.setDisplayedValues(puzz);
			populateEventsList(coalescedEvents);
			}
		});
	}

private void initPanel()
	{
//	setBorder(new TitledBorder("Solver"));
	setLayout(new BorderLayout(5, 5));
	
	JPanel north = new JPanel();
	north.setLayout(new BorderLayout(5, 5));
	add(north, BorderLayout.NORTH);

//	JPanel stPanel = new JPanel();
//	stPanel.setBorder(new TitledBorder("Status"));
//	north.add(stPanel, BorderLayout.NORTH);
	
	statusLbl = new JLabel("Application Ready");
	statusLbl.setBorder(new EmptyBorder(5, 10, 5, 5));
	north.add(statusLbl, BorderLayout.NORTH);
//	stPanel.add(statusLbl);

	final JSlider frameDelaySlider = new JSlider(JSlider.HORIZONTAL, 1, 70, (int)Math.sqrt(frameDelay) );
	frameDelaySlider.setBorder(new TitledBorder("Time delay (ms)"));
	frameDelaySlider.setPaintTicks(true);
	frameDelaySlider.setMinorTickSpacing(2);
	frameDelaySlider.setMajorTickSpacing(10);

	north.add(frameDelaySlider, BorderLayout.CENTER);

	eventList = new JList(new DefaultListModel());
	eventList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
	JScrollPane scroll = new JScrollPane(eventList);
	add(scroll, BorderLayout.CENTER);
	
	JPanel p = new JPanel();
	p.setLayout(new GridLayout(1, 0, 5, 5));
	
	playBtn = new JButton("Play");
	p.add(playBtn);

	pauseBtn = new JToggleButton("Pause");
	pauseBtn.setEnabled(false);
	p.add(pauseBtn);

	stopBtn = new JButton("Stop");
	stopBtn.setEnabled(false);
	p.add(stopBtn);
	
	stepBtn = new JButton("Step");
	p.add(stepBtn);
	
	add(p, BorderLayout.SOUTH);
	
	// add listeners
	
	frameDelaySlider.addChangeListener(new ChangeListener()
		{
		@Override
		public void stateChanged(ChangeEvent e)
			{
			int val = frameDelaySlider.getValue();
			frameDelay = val * val;
			}
		});
	
	eventListlistener = new ListSelectionListener()
		{
		@Override
		public void valueChanged(ListSelectionEvent e)
			{
			if ( e.getValueIsAdjusting() )
				return;
			
			if ( !enableListSelectionEvents )
				return;
			
//			stopSequence();
			gridGUI.clear(); 
			gridGUI.setInitialGrid(eventList.getSelectedIndex() == 0);
			
			SolutionEvent selected = (SolutionEvent)eventList.getSelectedValue();
			displaySelectedEvent(selected);
			player.setIndex(eventList.getSelectedIndex());
			}
		};
	eventList.addListSelectionListener(eventListlistener);
	
	// add btn listeners
	
	playBtn.addActionListener(new ActionListener()
		{
		@Override
		public void actionPerformed(ActionEvent e)
			{
			playSequence();
			}
		});
	
	pauseBtn.addActionListener(new ActionListener()
		{
		@Override
		public void actionPerformed(ActionEvent e)
			{
			pauseSequence(pauseBtn.isSelected());
			}
		});
	
	stopBtn.addActionListener(new ActionListener()
		{
		@Override
		public void actionPerformed(ActionEvent e)
			{
			stopSequence();
			}
		});
	
	stepBtn.addActionListener(new ActionListener()
		{
		@Override
		public void actionPerformed(ActionEvent e)
			{
			stepSequence();
			}
		});
	
	}

protected void stopSequence()
	{
	player.stop();

	playBtn.setEnabled(true);
	stepBtn.setEnabled(true);
	pauseBtn.setEnabled(false);
	stopBtn.setEnabled(false);
	}

protected void pauseSequence(boolean pause)
	{
	pauseBtn.setSelected(pause);
	stepBtn.setEnabled(pause);

	if ( pause )
		pauseBtn.setText("Resume");
	else
		pauseBtn.setText("Pause");

	
	player.pause(pause);
	}

protected void displaySelectedEvent(SolutionEvent selected)
	{
	int i = coalescedEvents.indexOf(selected);
	
	if ( i < 0 )
		return;
	
	Puzzle start;
	if ( i == 0 )
		start = seqBuf.getStartGrid();
	else
		start = coalescedEvents.get(i-1).getPuzzle();
	
	// set the grid to the start grid
	gridGUI.setDisplayedValues(start);
	}

protected void playSequence()
	{
	setSelectedIndex(0);
		
	playBtn.setEnabled(false);
	stepBtn.setEnabled(false);
	pauseBtn.setEnabled(true);
	stopBtn.setEnabled(true);

	player.play();
	}

protected void stepSequence()
	{
	player.singleStep();
	}

/**
 * Clears the solutions list
 */
void clear()
	{
	populateEventsList(new ArrayList<SolutionEvent>());
	}

private void populateEventsList(List<SolutionEvent> events)
	{
	eventList.removeListSelectionListener(eventListlistener);
	
	DefaultListModel model = (DefaultListModel)eventList.getModel();
	model.clear();
	
	for ( SolutionEvent e : events )
	    {
	    model.addElement(e);
	    }

	setSelectedIndex(0);
	eventList.addListSelectionListener(eventListlistener);
	
	if ( events.size() == 0 || sudokuSolver.getPuzzle().isComplete() )
		{
		eventList.setBackground(Color.WHITE);
		}
	else
		{
		eventList.setBackground(Color.PINK);
		}
	}

@Override
public void gridChanged(GridChangeEvent e) throws IllegalGridException
	{
	if ( !autoSolve )
		{
		return;
		}
	
	showSolution(new Puzzle(gridGUI.getDisplayedValues(), PuzzleDifficulty.UNKNOWN));
	
	updateStatusBar(sudokuSolver.getPuzzle());
	}


@Override
public void gridReplaced(Puzzle p) throws IllegalGridException
	{
	if ( !autoSolve )
		{
		updateStatusBar(p);
		return;
		}
	
	showSolution(p);
	updateStatusBar(sudokuSolver.getPuzzle());
	}

/**
 * Shows the solution for this puzzle
 * 
 * @param p the puzzle
 * @throws IllegalGridException
 */
void showSolution(Puzzle p) throws IllegalGridException
	{
	stopSequence();
	sudokuSolver.setPuzzle(p);
	sudokuSolver.compute();
	coalescedEvents = seqBuf.getCoalescedEvents();
	populateEventsList(coalescedEvents);
	}

private void updateStatusBar(Puzzle puzzle)
	{
	if ( !puzzle.isLegal() )
		{
		statusLbl.setText("The puzzle is not legal");
		}
	else if ( !puzzle.isComplete() && autoSolve )
		{
		statusLbl.setText("The puzzle is not solvable - try adding more values");
		}
	else
		{
		statusLbl.setText("Puzzle difficulty = "+puzzle.getDifficulty());
		}
	}

/**
 * Returns true if auto solve is turned on
 * 
 * @return the autoSolve state
 */
public boolean isAutoSolve()
	{
	return autoSolve;
	}

/**
 * Sets the auto solve state
 * 
 * @param state true to turn auto solve on
 */
public void setAutoSolve(boolean state)
	{
	this.autoSolve = state;
	
	if ( !autoSolve )
		statusLbl.setText("Auto solver disabled");
	}

private void selectEvent(SolutionEvent e)
    {
    eventList.setSelectedValue(e, true);
    }

/**
 * Gets the index of the currently selected solution event
 * 
 * @return the selected index
 */
int getSelectedIndex()
	{
	return eventList.getSelectedIndex();
	}

/**
 * Sets the index of the currently selected solution event
 * 
 * @param index the selected index
 */
void setSelectedIndex(int index)
	{
	eventList.setSelectedIndex(index);
	eventList.ensureIndexIsVisible(index);
	}

//======================================================================================================
/**
 * A Player for events lists. This runs on a background thread to free up the event thread
 * 
 * @author A.G.Docherty
 *
 */
class Player implements Runnable
	{
	private volatile boolean active;
	private volatile boolean singleStep;
	private volatile boolean pause;
	private Thread bkgThread;
	private volatile int index;
	private volatile boolean reset;
	
	public Player()
	    {
	    init();
	    }

	public void play()
	    {
	    index = 0;
	    reset = true;
	    pause = false;
	    singleStep = false;
	    bkgThread.interrupt();
	    }

	public void pause(boolean state)
	    {
	    singleStep = false;
	    pause = state;
	    bkgThread.interrupt();
	    }

	public void stop()
	    {
	    index = -1;
	    pause = true;
	    bkgThread.interrupt();
	    }

	public void singleStep()
	    {
	    singleStep = true;
	    pause = false;
	    bkgThread.interrupt();
	    }

	public void setIndex(int i)
	    {
	    reset = true;
	    index = i;
	    bkgThread.interrupt();
	    }

	public void init()
		{
		index = -1;
		active = true;
		singleStep = true;
		pause = true;
		bkgThread = new Thread(this, "Player");
		bkgThread.setDaemon(true);
		bkgThread.start();
		}
	
	public void term()
		{
		active = false;
		singleStep = false;
		
		bkgThread.interrupt();
		}
	
	@Override
    public void run()
		{
	    // play the sequence
		
		int i = 0;
		
		while ( active  )
			{
			gridGUI.setInitialGrid(false);
			enableListSelectionEvents = true; 
			reset = false;
			
			// if the index has been changed by the user jump to the new index now
			if ( index >= 0 )
				{
				i = index;
				index = -1;
				}
			

			// wait for the set time delay
			if ( !timeDelay() )
				{
				continue;
				}
			
			if ( i >= coalescedEvents.size() )
				continue;
			
			SolutionEvent e = coalescedEvents.get(i);
			runEvent(e);
			
			i++;
			if ( i >= coalescedEvents.size() )
				{
				// sequence finished 
				stopSequence();

		        // Remove any highlighting etc from the grid 
				EventQueue.invokeLater(new Runnable()
					{
					@Override
					public void run()
						{
						gridGUI.clear();
						}
					});
				
				}
			}
		
		
		}

	private void runEvent(final SolutionEvent e)
	    {
	    // update the screen: select event in event list
		EventQueue.invokeLater(new Runnable()
			{
			@Override
			public void run()
				{
				gridGUI.clear();
				enableListSelectionEvents = false; 
				selectEvent(e);
				enableListSelectionEvents = true; 
				}
			});

		if ( e.displayOrder() == DisplayOrder.EDITED_FIRST )
			{
	        // update the screen: select cells to be edited
			highlightSelectedCells(e);
	
	        // update the screen: select the referenced cells
		    if ( !e.getReferencedLocations().isEmpty() )
				{
				if ( !timeDelay() )
					return;
	
				highlightReferenceCells(e);
				}
			}
		else
			{
	        // update the screen: select the referenced cells
		    if ( !e.getReferencedLocations().isEmpty() )
				{
				highlightReferenceCells(e);

				if ( !timeDelay() )
					return;
				}
		    
	        // update the screen: select cells to be edited
			highlightSelectedCells(e);
			}

		if ( !timeDelay() )
			return;

        // update the screen: select the values to be removed from the selected cells
		EventQueue.invokeLater(new Runnable()
			{
			@Override
			public void run()
				{
				List<Point> loc = e.getLocations();
				
				gridGUI.setSelectedCells(loc, e.getRemovedValues());
//
//				if ( e.isResolved() )
//					gridGUI.setResolvedValues(loc, e.getGrid());
				}
			});

		if ( !timeDelay() )
			return;

        // update the screen: Remove the values from the button
		EventQueue.invokeLater(new Runnable()
			{
			@Override
			public void run()
				{
				List<Point> loc = e.getLocations();
				Cell[][] grid = e.getPuzzle().getGrid();
				
				for ( Point p : loc )
	                {
					gridGUI.setPossibleValues(p.x, p.y, grid[p.y][p.x]);
	                }
				}
			});
	    }

	/**
	 * Highlight the reference cells
	 * 
     * @param e the event
     */
    private void highlightReferenceCells(final SolutionEvent e)
	    {
        // update the screen: select the referenced cells
		EventQueue.invokeLater(new Runnable()
			{
			@Override
			public void run()
				{
				List<Point> locs = e.getReferencedLocations();
				gridGUI.setReferenced(locs, e.getReferencedValues());
				}
			});
	    }

	/**
     * @param e
     */
    private void highlightSelectedCells(final SolutionEvent e)
	    {
	    EventQueue.invokeLater(new Runnable()
			{
			@Override
			public void run()
				{
				List<Point> loc = e.getLocations();
				gridGUI.setSelectedCells(loc, null);
				}
			});
	    }

	private boolean timeDelay()
	    {
		do
			{
			try
	            {
	            Thread.sleep(frameDelay);
	            }
            catch ( InterruptedException exc )
	            {
	            // do nothing
	            }

            if ( !active )
            	return false;
            
            if ( reset )
            	{
            	reset = false;
            	return false;
            	}
			}
		while ( pause );
		
		if ( singleStep )
			pause = true;
		
	    return true;
	    }
	}
}
