/**
 *  Copyright (c) 2010 A.G.Docherty. All Rights Reserved.
 */

package uk.co.keang.sudoku;

import java.awt.Component;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintWriter;

import javax.swing.JFileChooser;
import javax.swing.filechooser.FileFilter;

import uk.co.keang.sudoku.event.IllegalGridException;


/**
 * Provides persistent storage and recovery of sudoku puzzles
 * 
 * @author A.G.Docherty
 *
 */
public class Storage
{
protected static final String SUDOKU_EXT = ".sdk";

/**
 * Saves this puzzle displaying a file chooser dialog defaulting to the users default directory
 *  
 * @param p the puzzle to save 
 * @param owner the gui owner or null
 * @return the file the puzzle was saved to or null if it wasn't saved
 * @throws IOException 
 */
public static File save(Puzzle p, Component owner) throws IOException
	{
	return save(p, owner, null);
	}

/**
 * Saves this puzzle displaying a file chooser dialog defaulting to the given directory
 *  
 * @param p the puzzle to save 
 * @param owner the gui owner or null
 * @param dir the directory or null. If a file is given, the files parent directory is used.
 * @return the file the puzzle was saved to or null if it wasn't saved
 * @throws IOException 
 */
public static File save(Puzzle p, Component owner, File dir) throws IOException
	{
	File saveTo = selectFile(owner, dir, "Save");
    save(p, saveTo);
    
    return saveTo;
	}

/**
 * Saves this puzzle to this file. If the file exists it will be overwritten.
 *  
 * @param p the puzzle to save 
 * @param file the file to save to
 * @throws IOException 
 */
public static void save(Puzzle p, File file) throws IOException
	{
	if ( file == null )
		return;
	
	PrintWriter w = new PrintWriter(file);
	
	try {
		p.write(w);
		}
	finally
		{
		w.close();
		}
	}

/**
 * Loads a puzzle, displaying a file chooser dialog defaulting to the users default directory
 *  
 * @param owner the gui owner or null
 * @return the puzzle 
 * @throws IOException 
 * @throws IllegalGridException 
 */
public static Puzzle load(Component owner) throws IOException, IllegalGridException
	{
	return load(owner, null);
	}

/**
 * Loads a puzzle, displaying a file chooser dialog defaulting to the given directory
 *  
 * @param owner the gui owner or null
 * @param dir the directory or null. If a file is given, the files parent directory is used.
 * @return the puzzle 
 * @throws IOException 
 * @throws IllegalGridException 
 */
public static Puzzle load(Component owner, File dir) throws IOException, IllegalGridException
	{
	File file = selectFile(owner, dir, "Load");
    Puzzle p = load(file);
    return p;
	}

/**
 * Loads the puzzle from this file. 
 *  
 * @param file the file to load from
 * @return the puzzle 
 * @throws IOException 
 * @throws IllegalGridException 
 */
public static Puzzle load(File file) throws IOException, IllegalGridException
	{
	if ( file == null )
		return null;
	
	BufferedReader r = new BufferedReader(new FileReader(file));
	
	Puzzle p = null;
	try 
		{
		p  = new Puzzle(r);
		}
	finally
		{
		r.close();
		}
		
	return p;
	}

/**
 * Displays a file chooser dialog
 * 
 * @param owner
 * @param dir
 * @return
 */
public static File selectFile(Component owner, File dir, String btnText)
	{
	JFileChooser fileChooser = new JFileChooser();
	
    fileChooser.setAcceptAllFileFilterUsed(false);
    fileChooser.setFileSelectionMode(JFileChooser.FILES_AND_DIRECTORIES);

    if ( dir != null )
    	{
	    if ( dir.isDirectory() )
	        fileChooser.setCurrentDirectory(dir);
	    else
	        {
	        fileChooser.setCurrentDirectory(dir.getParentFile());
	        }
    	}
    
    // set a filter for sudoku files and directories
    fileChooser.setFileFilter(new FileFilter() {
        @Override
        public boolean accept(File f)
            {
            if (f.isDirectory())
                return true;

            if ( f.getName() != null && f.getName().endsWith(SUDOKU_EXT) )
                {
                return true;
                }

            return false;
            }
        @Override
        public String getDescription()
            {
            return SUDOKU_EXT;
            }
        });

    int returnVal = fileChooser.showDialog(owner, btnText);

    if ( returnVal != JFileChooser.APPROVE_OPTION )
        {
        return null;
        }
    
    String file = fileChooser.getSelectedFile().getAbsolutePath();
    
    if ( !file.endsWith(SUDOKU_EXT) )
    	file = file + SUDOKU_EXT;
    
	return new File(file);
	}
}
