/**
 *  Copyright (c) 2010 A.G.Docherty. All Rights Reserved.
 */

package uk.co.keang.sudoku;

import java.util.ArrayList;
import java.util.List;

import uk.co.keang.sudoku.Puzzle.Cell;
import uk.co.keang.sudoku.Puzzle.PuzzleDifficulty;
import uk.co.keang.sudoku.event.IllegalGridException;
import uk.co.keang.sudoku.solver.AbstractSolver;
import uk.co.keang.sudoku.solver.AbstractSolver.SolverDifficulty;
import uk.co.keang.sudoku.solver.AssociatedPairsSolver;
import uk.co.keang.sudoku.solver.AsymmetricGroupsSolver;
import uk.co.keang.sudoku.solver.FixedLineSolver;
import uk.co.keang.sudoku.solver.HiddenGroupsSolver;
import uk.co.keang.sudoku.solver.KnownValueSolver;
import uk.co.keang.sudoku.solver.MatchingGroupsSolver;
import uk.co.keang.sudoku.solver.SplitGroupsSolver;
import uk.co.keang.sudoku.solver.ZeroChoiceSolver;

/**
 * A Sudoku puzzle solver
 * 
 * @author A.G.Docherty
 *
 */
public class SudokuSolver
{
public static final int SUB_GRID_SIZE = 3;
public static final int GRID_SIZE = SUB_GRID_SIZE * SUB_GRID_SIZE;

private final List<AbstractSolver> solvers = new ArrayList<AbstractSolver>();
private SequenceBuffer sequenceBuffer;
private final Puzzle puzzle;

public SudokuSolver()
	{
	this(null);
	}

public SudokuSolver(String[] args)
    {
    puzzle = new Puzzle();
    
    initSolvers();

    if ( args != null )
    	initArray(args);
    }

/**
 * init the solvers
 */
private void initSolvers()
	{
    solvers.add(new ZeroChoiceSolver(puzzle));
    solvers.add(new KnownValueSolver(puzzle));
    solvers.add(new MatchingGroupsSolver(puzzle));
    solvers.add(new FixedLineSolver(puzzle));
    solvers.add(new AsymmetricGroupsSolver(puzzle));
    solvers.add(new HiddenGroupsSolver(puzzle));
    solvers.add(new SplitGroupsSolver(puzzle));
    solvers.add(new AssociatedPairsSolver(puzzle));
	}

/**
 * Gets the puzzle
 * 
 * @return the puzzle
 */
public Puzzle getPuzzle()
	{
	return puzzle;
	}

/**
 * Sets the solvers puzzle grid to this puzzles grid
 * 
 * @param the puzzle
 */
public void setPuzzle(Puzzle p)
	{
	puzzle.setGrid(p.getGrid());
	}

/**
 * Checks the value is between 1 and GRID_SIZE inclusive
 * 
 * @param num
 * @return
 */
boolean isInBounds(int num)
	{
	if ( num < 1 || num > GRID_SIZE )
		{
		System.out.println("The number must be between 1 and "+GRID_SIZE+" inclusive");
		return false;
		}
	
	return true;
	}

/**
 * Runs the solvers
 * 
 * @return the difficulty where a higher number denotes a more difficult puzzle
 * @throws IllegalGridException 
 */
int compute() throws IllegalGridException
	{
	if ( sequenceBuffer != null )
		{
		sequenceBuffer.clear();
		sequenceBuffer.setStartGrid(puzzle);
		}
	
    int totalApplied = 0;
    int difficulty = 0;
    SolverDifficulty maxDif = SolverDifficulty.UNKNOWN;
	
	do  {
    	totalApplied = 0;
    	
		for ( int i = 0; i < solvers.size(); i++ )
			{
			AbstractSolver s = solvers.get(i);
			int difficultyFactor = s.getDifficultyFactor();
			int found = 0;

			found = s.findSolutions();

			if ( found > 0 )
				{
				SolverDifficulty d = s.getSolverDifficulty();
				if ( d.ordinal() > maxDif.ordinal() )
					maxDif = d;
				
				difficulty += found * difficultyFactor;
				totalApplied += found;

				// reset the solver sequence
				i = -1;
				}

			// if ( applied > 0 )
			// printDetail();
			}
    	
    	}
    while ( totalApplied != 0 );

    if ( !puzzle.isComplete() )
    	{
    	difficulty = -1;
    	}
    
    puzzle.setDifficulty(maxDif, difficulty);
//	System.out.println("Difficulty = "+difficulty);
    
    return difficulty;
	}

/**
 * Adds a solver listener to all solvers 
 *
 * @param listener - the listener to add
 */
public void addSolverListener(SolverListener listener)
	{
    for ( AbstractSolver s : solvers )
	    {
	    s.addSolverListener(listener);
	    }
	}

/**
 * Removes the solver listener from all solvers
 *
 * @param listener - the listener to remove
 */
public void removeSolverListener(SolverListener listener)
	{
    for ( AbstractSolver s : solvers )
	    {
	    s.removeSolverListener(listener);
	    }
	}

private void initArray(String[] args)
	{
	puzzle.clear();
	
	// if this is in 1 string split it
	if ( args.length == 1 )
		{
		char[] c = args[0].toCharArray();
		
		args = new String[c.length];
		
		for ( int i = 0; i < c.length; i++ )
	        {
	        args[i] = String.valueOf(c[i]);
	        }
		}
		
	// extract the values	
	for ( int i = 0; i < args.length; i++ )
	    {
	    int num;
	    
	    if ( !args[i].equals(".") && !args[i].equals(" ") )
	    	{
	    	num = Integer.parseInt(args[i]);
		    puzzle.set(i+1, num);
	    	}
	    }
	}

/**
 * Dumps a formatted view of possible values for each cell to the screen
 */
public void printDetail()
	{
    Cell[][] hOptions = puzzle.getGrid();

	StringBuffer sb = new StringBuffer();
	for ( int i = 0; i < (GRID_SIZE+2)*SUB_GRID_SIZE*SUB_GRID_SIZE+4; i++ )
	    {
	    sb.append("-");
	    }
	String line = sb.toString();
	
    for ( int i = 0; i < GRID_SIZE; i++ )
	    {
	    Cell[] x = hOptions[i];
	    
        if ( i % SUB_GRID_SIZE == 0 )
        	{
        	System.out.println(line);
        	}
        
	    for ( int j = 0; j < x.length; j++ )
	        {
	        if ( j % SUB_GRID_SIZE == 0 )
	        	System.out.print("|");

	        if ( x[j].size() == 1 )
	        	System.out.print("     "+x[j].iterator().next()+"     ");
	        else
	        	{
	        	for ( int j2 = (GRID_SIZE-hOptions[i][j].size())/2; j2 > 0; j2-- )
	                {
		        	System.out.print(" ");
	                }

	        	System.out.print("[");
	        	
	        	for ( Integer n : hOptions[i][j] )
	                {
	            	System.out.print(n);
	                }
	        	
	        	System.out.print("]");
        	
	        	for ( int j2 = (GRID_SIZE-hOptions[i][j].size()+1)/2; j2 > 0; j2-- )
	                {
		        	System.out.print(" ");
	                }
	        	}
	        
//	        System.out.print(" ");
	        }
	    
		System.out.println("|");
	    }
    
	System.out.println(line);
	}

/**
 * dumps a formatted view of resolved values to the screen
 */
void print()
	{
    Cell[][] hOptions = puzzle.getGrid();
	StringBuffer sb = new StringBuffer();
	for ( int i = 0; i < GRID_SIZE*2+4; i++ )
	    {
	    sb.append("-");
	    }
	String line = sb.toString();
	
    for ( int i = 0; i < GRID_SIZE; i++ )
	    {
	    Cell[] x = hOptions[i];
	    
        if ( i % SUB_GRID_SIZE == 0 )
        	{
        	System.out.println(line);
        	}
        
	    for ( int j = 0; j < x.length; j++ )
	        {
	        if ( j % SUB_GRID_SIZE == 0 )
	        	System.out.print("|");

	        if ( x[j].size() == 1 )
	        	System.out.print(x[j].iterator().next());
	        else
	        	System.out.print(".");

	        System.out.print(" ");
	        }
	    
		System.out.println("|");
	    }
	System.out.println(line);

//  Prints the numbers as a single line if values	
//	for ( Set<Integer>[] set : hOptions )
//	    {
//	    for ( Set<Integer> s : set )
//	        {
//	        System.out.print(s.iterator().next());
//	        }
//	    }
//	System.out.println();
	}

public SequenceBuffer getSequenceBuffer()
	{
	return sequenceBuffer;
	}

public void setSequenceBuffer(SequenceBuffer buf)
	{
	if ( sequenceBuffer != null )
		removeSolverListener(sequenceBuffer);
	
	sequenceBuffer = buf;
	addSolverListener(sequenceBuffer);
	}


//=========================================================================================================

public static void main(String[] args)
	{
	SudokuSolver s = new SudokuSolver();
	
	Generator generator = new Generator();
	Puzzle p = null;
	
	long time = System.currentTimeMillis();
	for ( int i = 0; i < 10; i++ )
	    {
	    p = generator.generatePuzzle(PuzzleDifficulty.MEDIUM, null);
	    System.out.println("Puzzle Generated");
	    s.setPuzzle(p);
	    
	    try
	        {
	        Thread.sleep(25);
	        }
        catch ( InterruptedException exc )
	        {
	        exc.printStackTrace();
	        }
        
//    	s.print();
	    }
	
	System.out.println("Time = "+(System.currentTimeMillis()-time));

	}
}
