/**
 *  Copyright (c) 2010 A.G.Docherty. All Rights Reserved.
 */

package uk.co.keang.sudoku.event;

import java.awt.Point;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.EventObject;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import uk.co.keang.sudoku.Puzzle;
import uk.co.keang.sudoku.Puzzle.Cell;
import uk.co.keang.sudoku.SudokuSolver;
import uk.co.keang.sudoku.solver.AbstractSolver;

/**
 * A solution event
 * 
 * @author A.G.Docherty
 *
 */
public class SolutionEvent extends EventObject implements Cloneable
{
private static final long serialVersionUID = 7991800393576086179L;
protected static String[] lblTextXAxis = new String[] { "A", "B", "C", "D", "E", "F", "G", "H", "I", "J", "K", "L", "M", "N", "O", "P" };

public enum DisplayOrder
	{
	EDITED_FIRST,
	REFERENCE_FIRST,
	}

public enum Grouping
	{
	ROW,
	COLUMN,
	BLOCK,
	GRID;
	}

private final String description;
private Cell removedValues;
private List<Point> locations;
private Puzzle puzzle;
private Cell[][] hGrid;
private List<Point> referenceLocations;
private final Grouping group;
private DisplayOrder displayOrder = DisplayOrder.EDITED_FIRST;

/**
 * A solver event for a changed cell
 * 
 * @param src  the solver that generated the event
 * @param msg  a descriptive message from the solver
 * @param grid  the grid after the proposed changes are applied
 * @param loc  the locations (0 based) of the cells that has been modified
 * @param removed  the values that been removed from this cells possible values list
 * @param from  a list of locations (0 based) of the cells used to determine which values can be removed 
 * @throws IllegalGridException 
 */
public SolutionEvent(AbstractSolver src, String msg, Puzzle grid, List<Point> loc, Cell removed, List<Point> from, Grouping g) throws IllegalGridException
	{
	super(src);

	description = msg;
	puzzle = grid;
	hGrid = puzzle.getGrid();
	group = g;
	
	if ( loc == null )
		locations = new ArrayList<Point>();
	else
		locations = new ArrayList<Point>(loc);
	
	if ( !isLegal() )
		throw new IllegalGridException(getIllegalValues());
	
	if ( removed != null )
		removedValues = new Cell(removed);
	else
		removedValues = new Cell();
	
	if ( from != null )
		referenceLocations = new ArrayList<Point>(from);
	else
		referenceLocations = new ArrayList<Point>();
	}

/**
 * true if this event change is legal
 * @return
 */
private boolean isLegal()
	{
	// check each location is legal
	for ( Point p : locations )
	    {
	    if ( !getIllegalValues(p).isEmpty() )
	    	return false;
	    }
	
	return true;
	}

/**
 * Gets any values in the row, column or block that are the duplicate of the values
 * at this events locations
 * 
 * @return the illegal value locations or null if none are found
 */
private List<Point> getIllegalValues()
	{
	List<Point> vals = new ArrayList<Point>();
	
	// check each location is legal
	for ( Point p : locations )
	    {
	    Set<Point> ret = getIllegalValues(p);
    	vals.addAll(ret);
	    }
	
	return vals;
	}

/**
 * Gets any values in the row, column or block that are the duplicate of the value at the given point
 * @param p
 * @return
 */
private Set<Point> getIllegalValues(Point p)
	{
	Set<Point> vals = new HashSet<Point>(); 
	
	// if there are multiple values for this cell it can't be illegal
	if ( hGrid[p.y][p.x].size() > 1 )
		return vals;
	
	// if there are no values for this cell it must be illegal
	if ( hGrid[p.y][p.x].size() == 0 )
		{
		vals.add(p);
		return vals;
		}
	
	// check this row for duplicates
	Cell[] row = hGrid[p.y];
	vals.addAll(getIllegalValues(row, p.x));
	
	// check this column for duplicates
	Cell[] col = puzzle.getColumn(p.x);
	vals.addAll(getIllegalValues(col, p.y));
	
	// check this block for duplicates
	Cell[] block = puzzle.getBlock(p.x, p.y);
	int index = p.x%SudokuSolver.SUB_GRID_SIZE + ((p.y%SudokuSolver.SUB_GRID_SIZE)*SudokuSolver.SUB_GRID_SIZE);
	vals.addAll(getIllegalValues(block, index));

	return vals;
	}

/**
 * Gets any values in the group that are the duplicate of the value at index
 * 
 * @param grp
 * @param index
 * @return
 */
private Set<Point> getIllegalValues(Cell[] grp, int index)
	{
	Set<Point> vals = new HashSet<Point>(); 
	
	int num = grp[index].iterator().next();
	
	for ( int i = 0; i < grp.length; i++ )
	    {
	    if ( i == index )
	    	continue;
	    
	    if ( grp[i].size() == 1 && grp[i].contains(num) )
	    	vals.add(puzzle.getLocation(grp[i]));
	    }

	if ( !vals.isEmpty() )
		{
    	vals.add(puzzle.getLocation(grp[index]));
		}
	
	return vals;
	}

/**
 * The values which are to be removed from this cell
 * @return the values
 */
public Cell getRemovedValues()
	{
	return removedValues;
	}

/**
 * The values which are to be highlighted in the referenced locations.
 * This implementation returns the removed values. Override this method
 * if you want different behaviour.
 * 
 * @return the values
 */
public Cell getReferencedValues()
	{
	return removedValues;
	}

/**
 * Gets the location of the squares that were used to create the change
 * 
 * @return
 */
public List<Point> getReferencedLocations()
	{
	return referenceLocations;
	}

/**
 * Gets the locations of the cells being modified or resolved
 * 
 * @return
 */
public List<Point> getLocations()
	{
	return locations;
	}

/**
 * The grid after any changes have been applied
 * 
 * @return
 */
public Puzzle getPuzzle()
	{
	return puzzle;
	}

/**
 * True if all of the location cells are resolved
 *
 * @return
 */
public boolean isResolved()
	{
	for ( Point p : locations )
	    {
	    if ( hGrid[p.y][p.x].size() > 1 )
	    	return false;
	    }
	
	return true;
	}

public String getDescription()
	{
	return description;
	}

public Grouping getGrouping()
	{
	return group;
	}

@Override
public SolutionEvent clone()
	{
	SolutionEvent clone = null;
	
    try
	    {
	    clone = (SolutionEvent)super.clone();
	    }
    catch ( CloneNotSupportedException exc )
	    {
	    // Can't happen as we implement Cloneable
	    exc.printStackTrace();
	    return null;
	    }

	clone.setPuzzle(puzzle.clone());
	clone.locations = new ArrayList<Point>(locations);
	clone.referenceLocations = new ArrayList<Point>(referenceLocations);
	clone.removedValues = new Cell(removedValues);
	
	return clone;
	}

/**
 * Returns true if the source and location are the same and if they both have change sets
 * 
 * @param e
 * @return
 */
public boolean canCoalesce(SolutionEvent e)
	{
	if ( e == null )
		return false;
	
	if ( source != e.source )
		return false;
	
	if ( !locations.equals(e.locations) )
		return false;

	if ( removedValues.equals(e.removedValues) )
		{
		return false;
		}
	
	return true;
	}

/**
 * Coalesces the given event with this one
 * 
 * @param e
 */
public void coalesce(SolutionEvent e)
	{
	removedValues.addAll(e.removedValues);
	
	for ( Point p : e.referenceLocations )
	    {
		if ( !referenceLocations.contains(p) )
			referenceLocations.add(p);
	    }
	
	for ( Point p : e.locations )
	    {
		if ( !locations.contains(p) )
			locations.add(p);
	    }
	
	setPuzzle(e.puzzle.clone());
	}

private void setPuzzle(Puzzle clone)
	{
	puzzle = clone;
	hGrid = puzzle.getGrid();
	}

@Override
public boolean equals(Object obj)
    {
    if ( obj == null || !getClass().equals(obj.getClass()) )
    	return false;
    
    SolutionEvent se = (SolutionEvent)obj;
    
    if ( !description.equals(se.description) )
    	return false;
    
    if ( !locations.equals(se.locations) )
    	return false;

    if ( !removedValues.equals(se.removedValues) )
    	{
    	return false;
    	}
    
    if ( !referenceLocations.equals(se.referenceLocations) )
       	{
       	return false;
       	}
       
    return true;
    }

@Override
public int hashCode()
    {
    return description.hashCode() + 
    	   (locations.hashCode()*7) + 
    	   (removedValues.hashCode()*13) +
    	   (referenceLocations.hashCode()*17);
    }

/**
 * The description of the event
 * 
 * @return
 */
@Override
public String toString()
	{
	String msg = getDescription();
	Cell removed = getRemovedValues();

	String locText = Puzzle.getLocationsAsText(getLocations());
	String numsText = null;
	
	msg = msg.trim();
	if ( !msg.endsWith(".") )
		msg += ".";
	
	if ( isResolved() )
		{
		Point p = locations.get(0);
		numsText = hGrid[p.y][p.x].iterator().next().toString();
		return MessageFormat.format("{0} The position {1} can only be a {2}", msg, locText, numsText);
		}

	numsText = Puzzle.getNumbersAsText(removed);
	return MessageFormat.format("{0} The position {1} can''t be a {2}", msg, locText, numsText);
	}

/**
 * Sets the display order for the event replay.
 * By default this is EDITED_FIRST.
 * 
 * @param order the order
 */
void setDisplayOrder(DisplayOrder order)
	{
	displayOrder = order;
	}

/**
 * Gets the display order for event replay. 
 * By default this is EDITED_FIRST.
 * 
 * @return the order
 */
public DisplayOrder displayOrder()
	{
	return displayOrder;
	}


}
