/**
 *  Copyright (c) 2010 A.G.Docherty. All Rights Reserved.
 */

package uk.co.keang.sudoku.solver;

import java.awt.Point;
import java.util.ArrayList;
import java.util.List;

import uk.co.keang.sudoku.Puzzle;
import uk.co.keang.sudoku.Puzzle.Cell;
import uk.co.keang.sudoku.SudokuSolver;
import uk.co.keang.sudoku.event.AsymmetricGroupsEvent;
import uk.co.keang.sudoku.event.IllegalGridException;
import uk.co.keang.sudoku.event.SolutionEvent;
import uk.co.keang.sudoku.event.SolutionEvent.Grouping;

/**
 * @author A.G.Docherty
 *
 * Provides resolution by looking for a square which has an option of 2 possible numbers and then
 * finding a pair of squares which contain 3 identical numbers of which 2 are the possible numbers. 
 * The third number can be removed from those values from the remaining empty squares 
 */
public class AsymmetricGroupsSolver extends AbstractSolver
{
public AsymmetricGroupsSolver(Puzzle grid)
    {
    super(grid, 50, SolverDifficulty.HARD);
    }

@Override
public int findSolutions() throws IllegalGridException
	{
	found = 0;

	for ( int i = 0; i < SudokuSolver.GRID_SIZE && found == 0; i++ )
		{
		found += resolveGroups(hGrid[i], "row", Grouping.ROW);
		}

	for ( int i = 0; i < SudokuSolver.GRID_SIZE && found == 0; i++ )
		{
		// create a set of the options column
		Cell[] col = puzzle.getColumn(i);

		found += resolveGroups(col, "column", Grouping.COLUMN);
		}

	for ( int i = 0; i < SudokuSolver.GRID_SIZE && found == 0; i++ )
		{
		// create a set of the options block
		Cell[] b = puzzle.getBlock(i);

		found += resolveGroups(b, "block", Grouping.BLOCK);
		}

	return found;
	}

private int resolveGroups(Cell[] opts, String line, Grouping direction) throws IllegalGridException
	{
	String msg = "Unequal Groups: Eliminate unequal matching triplets from the rest of the "+line;

	int f = 0;

	for ( int i = 0; i < SudokuSolver.GRID_SIZE; i++ )
		{
		// find a square containing 2 numbers
		if ( opts[i].size() == 2 )
			{
			// find a pair of squares containing 3 numbers of which 2 are the found numbers
			for ( int j = 0; j < SudokuSolver.GRID_SIZE; j++ )
				{
				// if this has 3 numbers and 2 are the found numbers
				if ( opts[j].size() == 3 && opts[j].containsAll(opts[i]) )
					{
					for ( int k = j+1; k < SudokuSolver.GRID_SIZE; k++ )
						{
						// if this is a matching group
						if ( opts[k].size() == 3 && opts[k].equals(opts[j]) )
							{
							// remove any references to these numbers from the rest of the line
							int nRemoved = 0;

							for ( int n = 0; n < SudokuSolver.GRID_SIZE; n++ )
								{
								// if this is one of the 3 squares or it is already resolved ignore it
								if ( n == i || n == j || n == k || opts[n].size() == 1 )
									continue;

								Cell removed = removeAllFromSet(opts[n], opts[j]);

								if ( removed.size() > 0 )
									{
									List<Point> coords = new ArrayList<Point>();
									coords.add(puzzle.getLocation(opts[i]));
									coords.add(puzzle.getLocation(opts[j]));
									coords.add(puzzle.getLocation(opts[k]));
									
									nRemoved += removed.size();
									firePossibilitiesReduced(hGrid, puzzle.getLocation(opts[n]), removed, coords, msg, direction);
									}
								}

							if ( nRemoved > 0 )
								f++;
							}
						}
					}
				}
			}
		}

	return f;
	}


/**
 * fires the possibilities reduced event to all listeners
 * @throws IllegalGridException 
 */
private void firePossibilitiesReduced(Cell[][] grid, Point loc, Cell removed, List<Point> from, String msg, Grouping direction) throws IllegalGridException
	{
	if ( !hasSolverListeners() )
		return;
	
	SolutionEvent e = new AsymmetricGroupsEvent(this, msg, puzzle.clone(), loc, removed, from, direction);
	fireSolutionEvent(e);
	}
}

