/**
 *  Copyright (c) 2010 A.G.Docherty. All Rights Reserved.
 */

package uk.co.keang.sudoku.solver;

import java.awt.Point;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Random;
import java.util.Stack;

import uk.co.keang.sudoku.Puzzle;
import uk.co.keang.sudoku.Puzzle.Cell;
import uk.co.keang.sudoku.SudokuSolver;
import uk.co.keang.sudoku.event.BruteForceEvent;
import uk.co.keang.sudoku.event.IllegalGridException;
import uk.co.keang.sudoku.event.SolutionEvent;
import uk.co.keang.sudoku.event.SolutionEvent.Grouping;


/**
 * @author A.G.Docherty
 *
 * Provides resolution by simply looking along row column and block for a missing value 
 */
public class BruteForceSolver extends AbstractSolver
{
@SuppressWarnings("unused")
private final Random random;
private static int recursionCount;

public BruteForceSolver(Puzzle grid, Random r)
    {
    super(grid, 1, SolverDifficulty.UNKNOWN);
    random = r;
    }

@Override
public int findSolutions() throws IllegalGridException
	{
	found = 0;
	
	int count = puzzle.countKnownValues();
	
	// FIXME sometimes this goes recursive too many times before finding a solution
	// the attempt below does not work either
	// most puzzles solve in under 300 cycles but some take 17000 and some appear to never solve.
	
	recursionCount = 0;
	
//	Puzzle p = puzzle.clone();
//	fillGrid(p, new Point(0,0));
	
	int rotation = 0;
	Puzzle p = null;
	
	for ( ; rotation < 4; rotation++ )
	    {
		recursionCount = 0;
		p = puzzle;
		if ( fillGrid(p, getNextLocation(p, new Point(-1,0))) )
			{
//			System.out.println("Solved - recursion = "+recursionCount);
			break;
			}
	    
//		System.out.println("Rotate "+(rotation+1));
		puzzle.rotate();
	    }
	
	if ( rotation == 4 )
		throw new IllegalGridException(null);
	
	while ( rotation > 0 && rotation < 4 )
		{
//		System.out.println("Complete rotation "+(rotation+1));
		p.rotate();
		rotation++;
		}
	
	found = p.countKnownValues() - count;
	
	if ( found > 0 )
		{
//		puzzle.setGrid(p.getGrid());
		firePossibilitiesReduced(hGrid);
		}
	
	return found;
	}

/**
 * Fills this puzzle grid point with a random numbers to make a legal cell. 
 * This uses a stack to allow backtracking.  
 * This is an alternative approach to recursion.
 * 
 * @param p
 * @param loc
 * @return
 */
@SuppressWarnings("unused")
private boolean fillGrid2(Puzzle p, Point loc)
	{
	recursionCount = 0;
	
	Stack<SolverState> stack = new Stack<SolverState>();

	List<Integer> possible = getRandomisedNumbers(p.getPossibleValues(loc));
	SolverState state = new SolverState(loc, possible);
	
	do {
		Cell cell = p.getGrid()[state.loc.y][state.loc.x];
	
		if ( state.possible.size() > 0 )
		    {
		    Integer i = state.possible.remove(state.possible.size()-1);
		    
		    // there's little chance of solving this puzzle
			if ( recursionCount > 1000 )
				{
				System.out.println("Very Recursive = "+recursionCount);
	//			return false;
				}
			
		    cell.clear();
		    cell.add(i);
		    
		    // get next location
		    Point nextloc = getNextLocation(p, loc);
			if ( nextloc == null )
		    	{
		    	// finished
		    	return true;
		    	}
			
			// add this item to the stack
            recursionCount++;
            stack.add(state);
            List<Integer> possibleNums = getRandomisedNumbers(p.getPossibleValues(nextloc));
            state = new SolverState(nextloc, possibleNums);
            continue;
		    }
	
		// remove any incorrect values
		cell.clear();

		// pop the last item off the stack
    	state = stack.pop();
		}
	while ( true );
	}

/**
 * Fills this puzzle grid point with a random numbers to make a legal cell. 
 * This recursively calls itself to complete the grid from this point onwards.  
 * 
 * @param p
 * @param loc
 * @return
 */
private boolean fillGrid(Puzzle p, Point loc)
	{
	recursionCount++;
	
	List<Integer> possible = getRandomisedNumbers(p.getPossibleValues(loc));
	
	Cell cell = p.getGrid()[loc.y][loc.x];

	for ( Integer i : possible )
	    {
	    // there's little chance of solving this puzzle
		if ( recursionCount > 1000 )
			{
//			System.out.println("Very Recursive = "+recursionCount);
//			recursionCount = 0;
//			return false;
			}
		
	    cell.clear();
	    cell.add(i);
	    
	    // get next location
	    Point nextloc = getNextLocation(p, loc);
	    if ( nextloc == null )
	    	return true;
	    
	    boolean ok = fillGrid(p, nextloc);
	    if ( ok )
	    	return ok;
	    }

	// remove incorrect value
	cell.clear();
	
	return false;
	}

/**
 * Gets the next unknown location or null if the end of the grid has been reached.
 * Locations increment from left to right and then top to bottom.
 * 
 * @param loc the current location
 * @return the next unknown location or null if there are no more locations
 */
private Point getNextLocation(Puzzle p, Point loc)
	{
    Point nextLoc = (Point)loc.clone();
    nextLoc.x++;
    if ( nextLoc.x >= SudokuSolver.GRID_SIZE )
    	{
    	nextLoc.x = 0;
    	nextLoc.y++;
    	
    	if ( nextLoc.y >= SudokuSolver.GRID_SIZE )
    		return null;
    	}
    
    if ( p.getGrid()[nextLoc.y][nextLoc.x].size() == 1 )
    	return getNextLocation(p, nextLoc);
    
    return nextLoc;
	}

/**
 * get a randomised sequence of these numbers
 * 
 * @param nums the numbers to randomise
 * @return
 */
private List<Integer> getRandomisedNumbers(Cell nums)
	{
	List<Integer> randomised = new ArrayList<Integer>(nums.getCollection());
	
	if ( randomised.size() < 2 )
		return randomised;
	
	// this should be reinstated as it's a better random but the alternative method
	// is deterministic for performance testing purposes
	Collections.shuffle(randomised);
	
//	for ( int i = 0; i < randomised.size(); i++ )
//	    {
//	    int j = random.nextInt(randomised.size()-1);
//	    
//	    if ( j != i )
//	    	{
//	    	Integer elem = randomised.remove(i);
//	    	randomised.add(j, elem);
//	    	}
//	    }
	
	return randomised;
	}

/**
 * Fills the block index with a random set of numbers. These numbers are legal within this block
 * but may not be legal when compared with other blocks
 * 
 * @param blockIndex - 0n based block number where blocks are number left to right and then top to bottom
 * @return the number of numbers added
 */
public int fillBlock(int blockIndex)
	{
	Cell[] block = null;
	block = puzzle.getBlock(blockIndex);
	
	List<Integer> nums = getRandomisedNumbers(Puzzle.allPossibleNumbers);

	for ( int i = 0; i < block.length; i++ )
	    {
	    block[i].clear();
	    block[i].add(nums.get(i));
	    }
	
	return block.length;
	}

/**
 * fires the possibilities reduced event to all listeners
 * @throws IllegalGridException 
 */
private void firePossibilitiesReduced(Cell[][] grid) throws IllegalGridException
	{
	if ( !hasSolverListeners() )
		return;
	
	String msg = "Brute Force: cell value resolved.";
	
	List<Point> locs = new ArrayList<Point>();
	for ( int y = 0; y < grid.length; y++ )
	    {
	    for ( int x = 0; x < grid[y].length; x++ )
	        {
	        locs.add(new Point(x, y));
	        }
	    }
	
	SolutionEvent e = new BruteForceEvent(this, msg, puzzle.clone(), locs, Grouping.GRID);
	fireSolutionEvent(e);
	}

private class SolverState
{

private final Point loc;
private final List<Integer> possible;

public SolverState(Point nextloc, List<Integer> possibleNums)
	{
	loc = nextloc;
	possible = possibleNums;
	}

}
}

