/**
 *  Copyright (c) 2010 A.G.Docherty. All Rights Reserved.
 */

package uk.co.keang.sudoku.solver;

import java.awt.Point;
import java.util.ArrayList;
import java.util.List;

import uk.co.keang.sudoku.Puzzle;
import uk.co.keang.sudoku.Puzzle.Cell;
import uk.co.keang.sudoku.SudokuSolver;
import uk.co.keang.sudoku.event.IllegalGridException;
import uk.co.keang.sudoku.event.MatchingGroupEvent;
import uk.co.keang.sudoku.event.SolutionEvent;
import uk.co.keang.sudoku.event.SolutionEvent.Grouping;

/**
 * @author A.G.Docherty
 *
/**
 * Provides resolution by looking for pairs, triplets etc of possible values and if found
 * by removing those values from the remaining empty squares 
 */
public class MatchingGroupsSolver extends AbstractSolver
{

public MatchingGroupsSolver(Puzzle grid)
    {
    super(grid, 20, SolverDifficulty.MEDIUM);
    }

private final String[] type = new String[] {"Singles", "Pairs", "Triplets", "Quadruplets"};
@Override
public int findSolutions() throws IllegalGridException
	{
	found = 0;

	for ( int j = 2; j < 5; j++ )
		{
		for ( int i = 0; i < SudokuSolver.GRID_SIZE && found == 0; i++ )
			{
			found += resolveGroups(hGrid[i], j, "row", Grouping.ROW);
			}

		for ( int i = 0; i < SudokuSolver.GRID_SIZE && found == 0; i++ )
			{
			// create a set of the options column
			Cell[] col = puzzle.getColumn(i);

			found += resolveGroups(col, j, "column", Grouping.COLUMN);
			}

		for ( int i = 0; i < SudokuSolver.GRID_SIZE && found == 0; i++ )
			{
			// create a set of the options block
			Cell[] col = puzzle.getBlock(i);

			found += resolveGroups(col, j, "block", Grouping.BLOCK);
			}
		}

	return found;
	}

private int resolveGroups(Cell[] opts, int groupSize, String line, Grouping direction) throws IllegalGridException
	{
	int f = 0;
	List<Point> coords = new ArrayList<Point>();

	int nFreq = 0;
	
	// if there's not at least n groups of n exit now
	for ( int i = 0; i < opts.length; i++ )
	    {
	    if ( opts[i].size() == groupSize )
	    	{
	    	nFreq++;
	    	}
	    }
	
	if ( nFreq < groupSize )
		return 0;
	
	for ( int i = 0; i < SudokuSolver.GRID_SIZE-groupSize+1; i++ )
		{
		coords.clear();
		coords.add(puzzle.getLocation(opts[i]));
		
		int matches = 0;
		for ( int j = i+1; j < SudokuSolver.GRID_SIZE; j++ )
			{
			// if this are a matching group
			if ( opts[i].size() == groupSize && opts[i].equals(opts[j]) )
				{
				coords.add(puzzle.getLocation(opts[j]));
				matches++;
				}
			}

		// remove any references to these numbers from the rest of the line
		if ( matches == groupSize - 1 )
			{
			int nRemoved = 0;

			String msg = "Matching "+type[groupSize-1]+": Eliminating matching values from the rest of the "+line;

			for ( int j = 0; j < SudokuSolver.GRID_SIZE; j++ )
				{
				if ( opts[j].equals(opts[i]) )
					continue;

				if ( opts[j].size() == 1 )
					continue;

				Cell removed = removeAllFromSet(opts[j], opts[i]);

				if ( removed.size() > 0 )
					{
					nRemoved += removed.size();
					firePossibilitiesReduced(hGrid, puzzle.getLocation(opts[j]), removed, coords, msg, direction);
					}
				}

			if ( nRemoved > 0 )
				f++;
			}

		}

	return f;
	}

/**
 * fires the possibilities reduced event to all listeners
 * @throws IllegalGridException 
 */
private void firePossibilitiesReduced(Cell[][] grid, Point loc, Cell removed, List<Point> from, String msg, Grouping direction) throws IllegalGridException
	{
	if ( !hasSolverListeners() )
		return;
	
	SolutionEvent e = new MatchingGroupEvent(this, msg, puzzle.clone(), loc, removed, from, direction);
	fireSolutionEvent(e);
	}
}

