/**
 *  Copyright (c) 2010 A.G.Docherty. All Rights Reserved.
 */

package uk.co.keang.sudoku.solver;

import java.awt.Point;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import uk.co.keang.sudoku.Puzzle;
import uk.co.keang.sudoku.Puzzle.Cell;
import uk.co.keang.sudoku.SudokuSolver;
import uk.co.keang.sudoku.event.IllegalGridException;
import uk.co.keang.sudoku.event.SolutionEvent;
import uk.co.keang.sudoku.event.SolutionEvent.Grouping;
import uk.co.keang.sudoku.event.SplitGroupsEvent;

/**
 * Provides resolution by looking for a triplet that only contain 3 common numbers
 * (they do not each need to contain all of the common numbers but must not contain any other numbers)
 * these numbers can be removed from all other cells in the group.
 * This also tests for quads of 4 common numbers
 * 
 * @author A.G.Docherty
 *
 */
public class SplitGroupsSolver extends AbstractSolver
{
private final String[] type = new String[] {"single", "pair", "triplet", "quadruplet"};

public SplitGroupsSolver(Puzzle grid)
    {
    super(grid, 40, SolverDifficulty.HARD);
    }

@Override
public int findSolutions() throws IllegalGridException
	{
	found = 0;

	for ( int i = 0; i < SudokuSolver.GRID_SIZE && found == 0; i++ )
		{
		found += resolveSplitGroups(hGrid[i], Grouping.ROW);
		}

	for ( int i = 0; i < SudokuSolver.GRID_SIZE && found == 0; i++ )
		{
		// create a set of the options column
		Cell[] col = puzzle.getColumn(i);

		found += resolveSplitGroups(col, Grouping.COLUMN);
		}

	for ( int i = 0; i < SudokuSolver.GRID_SIZE && found == 0; i++ )
		{
		// create a set of the options block
		Cell[] col = puzzle.getBlock(i);

		found += resolveSplitGroups(col, Grouping.BLOCK);
		}

	return found;
	}

private int resolveSplitGroups(Cell[] sets, Grouping direction) throws IllegalGridException
	{
	int f = 0;
	int[] occur = Puzzle.getFrequencyOfPossibilities(sets);

	List<Integer> matches = new ArrayList<Integer>();
	
	for ( int n = 4; n > 2 && f == 0; n-- )
	    {
	    // if no numbers occur more than n times there is nothing to remove
	    if ( !hasFrequencies(occur, n+1) )
	    	continue;
	    
		matches.clear();
		List<Cell> cells = new ArrayList<Cell>();
		
		// find the cells that contain n or less numbers
		for ( Cell s : sets )
	        {
	        int size = s.size();
	        if ( size > 1 && size <= n )
	        	{
	        	cells.add(s);
	        	}
	        }
		
		// if there is not at least n cells then exit now
		if ( cells.size() < n )
			continue;
		
		// the set must contain at least 1 number with a frequency greater than n
		for ( int i = 0; i < occur.length; i++ )
	        {
	        if ( occur[i] > 1 )
	        	{
				// add n of these cells together to produce a set containing exactly n numbers
				// for each group found remove these numbers from the rest of group 
				List<Cell> common = getCommonSets(cells, i+1, n);
				
				if ( common == null )
					continue;
				
				// remove any extra value from the sets
				int ret = removeExtraValues(sets, common, n, direction);
				
				if ( ret > 0 )
					{
					f += ret;
					
					// remove any sets that now only have 1 possible value
					Iterator<Cell> it = common.iterator();
					while ( it.hasNext() )
						{
						Cell s = it.next();
						
						if ( s.size() == 1 )
							it.remove();
						}
					}
	        	}
	        }
		}
	
	return f;
	}
	
/**
 * Check to see if any of the frequencies are at least n 
 * 
 * @param freq
 * @param n
 * @return
 */
private boolean hasFrequencies(int[] freq, int n)
	{
    for ( int i = 0; i < freq.length; i++ )
        {
        if ( freq[i] >= n )
        	return true;
        }
    
	return false;
	}

/**
 * Removes these values for any set that contains values other than those in common
 * 
 * @param sets
 * @param common
 * @param n 
 * @throws IllegalGridException 
 */
private int removeExtraValues(Cell[] group, List<Cell> common, int n, Grouping direction) throws IllegalGridException
	{
	String msg = "Split Group: Common numbers in a "+type[n-1]+" can't be anywhere else in this "+direction.toString().toLowerCase()+".";

	int f = 0;
	
    Cell remove = new Cell();
	List<Point> coords = new ArrayList<Point>();
	
	for ( Cell s : common )
	    {
		coords.add(puzzle.getLocation(s));
	    remove.addAll(s);
	    }
	
	for ( int i = 0; i < group.length; i++ )
	    {
	    // if this set is one of the matching sets ignore it
	    if ( !common.contains(group[i]) )
	    	{
			Cell removed = removeAllFromSet(group[i], remove);

			if ( removed.size() > 0 )
				{
				f += removed.size();
				firePossibilitiesReduced(hGrid, puzzle.getLocation(group[i]), removed, coords, msg, direction);
				}
	    	}
		}
	
	return f;
	}

/**
 * Gets n sets that just contain a subset of the same n numbers (one of which must be num)
 * and remove them from the list
 * 
 * @param locs
 * @param n
 * @param n2 
 * @return
 */
private List<Cell> getCommonSets(List<Cell> cells, int num, int n)
	{
	// get all the <= n sized cells that this number
	List<Cell> list = new ArrayList<Cell>();

	for ( Cell s : cells )
	    {
	    if ( s.contains(num) )
	    	list.add(s);
	    }

	// the number must occur in at least 3 cells
	if ( list.size() < 3 )
		return null;
	
	Cell set = new Cell();
	
	// try each combination
	for ( int k = 0; k < list.size(); k++ )
		{
		// find n sets that contain this number
		for ( int i = k; i < list.size(); i++ )
		    {
		    set.addAll(list.get(i));
	
		    // if this set has the correct number of digits
		    if ( set.size() == n )
		    	{
				List<Cell> foundSets = new ArrayList<Cell>();
			    
		    	// find n-1 other sets containing only these numbers
		    	for ( int j = 0; j < cells.size(); j++ )
		            {
		            Cell s = cells.get(j);
		            
		            if ( set.containsAll(s) )
		            	{
		            	foundSets.add(s);
		            	}
		            }
		    	
		    	if ( foundSets.size() == n )
		    		return foundSets;
		    	}
		    }
		
		// rotate the list
		list.add(list.remove(0));
		set.clear();
		}
	
	return null;
	}


/**
 * fires the possibilities reduced event to all listeners
 * @throws IllegalGridException 
 */
private void firePossibilitiesReduced(Cell[][] grid, Point loc, Cell removed, List<Point> from, String msg, Grouping direction) throws IllegalGridException
	{
	if ( !hasSolverListeners() )
		return;
	
	SolutionEvent e = new SplitGroupsEvent(this, msg, puzzle.clone(), loc, removed, from, direction);
	fireSolutionEvent(e);
	}

}
