/**
 *  Copyright (c) 2010 A.G.Docherty. All Rights Reserved.
 */

package uk.co.keang.sudoku.solver;

import static org.junit.Assert.assertTrue;

import java.util.Arrays;

import org.junit.Before;
import org.junit.Test;

import uk.co.keang.sudoku.Puzzle;
import uk.co.keang.sudoku.Puzzle.Cell;
import uk.co.keang.sudoku.SudokuSolver;
import uk.co.keang.sudoku.event.IllegalGridException;

/**
 * @author A.G.Docherty
 * If there is N cells (2,3 or 4) containing only up to N common numbers, then all these numbers 
 * can be excluded from all other cells.
 */
public class SplitGroupsSolverTest
{

private Cell[][] grid;
private final String[] args = new String[] { "294135876813746925657298143926813754735964281148572369579621438382459617461387592" };
private SudokuSolver sudoku;
private Puzzle puzzle;

/**
 * @throws java.lang.Exception
 */
@Before
public void setUp() throws Exception
	{
	sudoku = new SudokuSolver(args);
	puzzle = sudoku.getPuzzle();
	grid = puzzle.getGrid();
	}

/**
 * Test method for {@link uk.co.keang.sudoku.solver.ZeroChoiceSolver#findSolutions()}.
 * @throws IllegalGridException 
 */
@Test
public void testFindSolutionsRows() throws IllegalGridException
	{
	puzzle.set(0, 1, new Cell(Arrays.asList(2, 4, 3)));
	puzzle.set(2, 1, new Cell(Arrays.asList(8, 4, 5)));
	puzzle.set(4, 1, new Cell(Arrays.asList(2, 3)));
	puzzle.set(6, 1, new Cell(Arrays.asList(4, 6, 8)));
	puzzle.set(7, 1, new Cell(Arrays.asList(2, 4)));
	puzzle.set(8, 1, new Cell(Arrays.asList(2, 4, 8, 6, 5)));

//	sudoku.printDetail();
	
	assertTrue(grid[1][0].size() > 1);
	assertTrue(grid[1][2].size() > 1);
	assertTrue(grid[1][4].size() > 1);
	assertTrue(grid[1][6].size() > 1);
	assertTrue(grid[1][7].size() > 1);
	assertTrue(grid[1][8].size() > 1);
	
	SplitGroupsSolver solver = new SplitGroupsSolver(puzzle);
	solver.findSolutions();

//	sudoku.printDetail();

	assertTrue(grid[1][0].size() == 3);
	assertTrue(grid[1][2].size() == 2);
	assertTrue(grid[1][4].size() == 2);
	assertTrue(grid[1][6].size() == 2);
	assertTrue(grid[1][7].size() == 2);
	assertTrue(grid[1][8].size() == 3);
	}

/**
 * Test method for {@link uk.co.keang.sudoku.solver.ZeroChoiceSolver#findSolutions()}.
 * @throws IllegalGridException 
 */
@Test
public void testFindSolutionsRows2() throws IllegalGridException
	{
	puzzle.set(0, 1, new Cell(Arrays.asList(4, 6, 8)));
	puzzle.set(2, 1, new Cell(Arrays.asList(2, 3)));
	puzzle.set(4, 1, new Cell(Arrays.asList(8, 4)));
	puzzle.set(5, 1, new Cell(Arrays.asList(3, 4)));
	puzzle.set(6, 1, new Cell(Arrays.asList(2, 8, 4, 5)));
	puzzle.set(7, 1, new Cell(Arrays.asList(2, 4)));
	puzzle.set(8, 1, new Cell(Arrays.asList(2, 4, 8, 6, 5)));

//	sudoku.printDetail();
	
	assertTrue(grid[1][0].size() > 1);
	assertTrue(grid[1][2].size() > 1);
	assertTrue(grid[1][4].size() > 1);
	assertTrue(grid[1][5].size() > 1);
	assertTrue(grid[1][6].size() > 1);
	assertTrue(grid[1][7].size() > 1);
	assertTrue(grid[1][8].size() > 1);
	
	SplitGroupsSolver solver = new SplitGroupsSolver(puzzle);
	solver.findSolutions();

//	sudoku.printDetail();

	assertTrue(grid[1][0].size() == 2);
	assertTrue(grid[1][2].size() == 2);
	assertTrue(grid[1][4].size() == 1);
	assertTrue(grid[1][5].size() == 2);
	assertTrue(grid[1][6].size() == 2);
	assertTrue(grid[1][7].size() == 2);
	assertTrue(grid[1][8].size() == 3);
	}

/**
 * Test method for {@link uk.co.keang.sudoku.solver.ZeroChoiceSolver#findSolutions()}.
 * @throws IllegalGridException 
 */
@Test
public void testFindSolutionsColumns() throws IllegalGridException
	{
	puzzle.set(1, 0, new Cell(Arrays.asList(9, 3)));
	puzzle.set(1, 2, new Cell(Arrays.asList(4, 8, 7, 5)));
	puzzle.set(1, 4, new Cell(Arrays.asList(9, 7, 3)));
	puzzle.set(1, 6, new Cell(Arrays.asList(7, 4)));
	puzzle.set(1, 7, new Cell(Arrays.asList(7, 9)));
	puzzle.set(1, 8, new Cell(Arrays.asList(3, 7, 8, 6, 5)));

//	sudoku.printDetail();
	
	assertTrue(grid[0][1].size() > 1);
	assertTrue(grid[2][1].size() > 1);
	assertTrue(grid[4][1].size() > 1);
	assertTrue(grid[6][1].size() > 1);
	assertTrue(grid[7][1].size() > 1);
	assertTrue(grid[8][1].size() > 1);
	
	SplitGroupsSolver solver = new SplitGroupsSolver(puzzle);
	solver.findSolutions();

//	sudoku.printDetail();

	assertTrue(grid[0][1].size() == 2);
	assertTrue(grid[2][1].size() == 3);
	assertTrue(grid[4][1].size() == 3);
	assertTrue(grid[6][1].size() == 1);
	assertTrue(grid[7][1].size() == 2);
	assertTrue(grid[8][1].size() == 3);
	}

/**
 * Test method for {@link uk.co.keang.sudoku.solver.ZeroChoiceSolver#findSolutions()}.
 * @throws IllegalGridException 
 */
@Test
public void testFindSolutionsBlocks() throws IllegalGridException
	{
	// triplet of 2,3,4
	puzzle.set(0, 0, new Cell(Arrays.asList(2, 4, 3)));
	puzzle.set(2, 0, new Cell(Arrays.asList(1, 4, 6)));
	puzzle.set(1, 1, new Cell(Arrays.asList(2, 3)));
	puzzle.set(2, 1, new Cell(Arrays.asList(4, 8, 6)));
	puzzle.set(0, 2, new Cell(Arrays.asList(2, 4)));
	puzzle.set(2, 2, new Cell(Arrays.asList(2, 4, 1, 6, 9)));

//	sudoku.printDetail();
	
	assertTrue(grid[0][0].size() > 1);
	assertTrue(grid[0][2].size() > 1);
	assertTrue(grid[1][1].size() > 1);
	assertTrue(grid[1][2].size() > 1);
	assertTrue(grid[2][0].size() > 1);
	assertTrue(grid[2][2].size() > 1);
	
	SplitGroupsSolver solver = new SplitGroupsSolver(puzzle);
	solver.findSolutions();

//	sudoku.printDetail();

	assertTrue(grid[0][0].size() == 3);
	assertTrue(grid[0][2].size() == 2);
	assertTrue(grid[1][1].size() == 2);
	assertTrue(grid[1][2].size() == 2);
	assertTrue(grid[2][0].size() == 2);
	assertTrue(grid[2][2].size() == 3);
	}

/**
 * Test method for {@link uk.co.keang.sudoku.solver.ZeroChoiceSolver#findSolutions()}.
 * @throws IllegalGridException 
 */
@Test
public void testFindSolutionsRowsQuad() throws IllegalGridException
	{
	// finds a quad value of 2,3,4,8 
	puzzle.set(0, 1, new Cell(Arrays.asList(2, 4, 3)));
	puzzle.set(2, 1, new Cell(Arrays.asList(8, 4, 2)));
	puzzle.set(4, 1, new Cell(Arrays.asList(2, 3)));
	puzzle.set(6, 1, new Cell(Arrays.asList(4, 6)));
	puzzle.set(7, 1, new Cell(Arrays.asList(2, 4)));
	puzzle.set(8, 1, new Cell(Arrays.asList(2, 4, 8, 6, 5)));

	sudoku.printDetail();
	
	assertTrue(grid[1][0].size() > 1);
	assertTrue(grid[1][2].size() > 1);
	assertTrue(grid[1][4].size() > 1);
	assertTrue(grid[1][6].size() > 1);
	assertTrue(grid[1][7].size() > 1);
	assertTrue(grid[1][8].size() > 1);
	
	SplitGroupsSolver solver = new SplitGroupsSolver(puzzle);
	solver.findSolutions();

	sudoku.printDetail();

	assertTrue(grid[1][0].size() == 3);
	assertTrue(grid[1][2].size() == 3);
	assertTrue(grid[1][4].size() == 2);
	assertTrue(grid[1][6].size() == 1);
	assertTrue(grid[1][7].size() == 2);
	assertTrue(grid[1][8].size() == 2);

	// find triplet of 2,3,4
	solver.findSolutions();

	sudoku.printDetail();

	assertTrue(grid[1][0].size() == 3);
	assertTrue(grid[1][2].size() == 1);
	assertTrue(grid[1][4].size() == 2);
	assertTrue(grid[1][6].size() == 1);
	assertTrue(grid[1][7].size() == 2);
	assertTrue(grid[1][8].size() == 2);
	}


}
