/**
 *  Copyright (c) 2010 A.G.Docherty. All Rights Reserved.
 */

package uk.co.keang.sudoku.solver;

import java.awt.Point;
import java.util.ArrayList;
import java.util.List;

import uk.co.keang.sudoku.Puzzle;
import uk.co.keang.sudoku.Puzzle.Cell;
import uk.co.keang.sudoku.SudokuSolver;
import uk.co.keang.sudoku.event.IllegalGridException;
import uk.co.keang.sudoku.event.SolutionEvent;
import uk.co.keang.sudoku.event.SolutionEvent.Grouping;
import uk.co.keang.sudoku.event.ZeroChoiceEvent;

/**
 * Provides resolution by looking for unresolved squares for lines where a number can only go 
 * in one square 
 * 
 * Because of the basic nature of this solver it repeatedly runs until no more solutions are found
 *    
 * @author A.G.Docherty
 */
public class ZeroChoiceSolver extends AbstractSolver
{
public ZeroChoiceSolver(Puzzle grid)
    {
    super(grid, 1, SolverDifficulty.BEGINNER);
    }

@Override
public int findSolutions() throws IllegalGridException
	{
	found = 0;

	int f = 0;
	do 
		{
		f = findOnlyPossibleSquare();
		found += f;
		}
	while ( f > 0 );

	return found;
	}

private int findOnlyPossibleSquare() throws IllegalGridException
	{
	int f = 0;

	for ( int i = 0; i < SudokuSolver.GRID_SIZE; i++ )
		{
		f += resolveUniquePositions(hGrid[i], Grouping.ROW);
		}

	for ( int i = 0; i < SudokuSolver.GRID_SIZE; i++ )
		{
		// create a set of the options column
		Cell[] col = puzzle.getColumn(i);

		f += resolveUniquePositions(col, Grouping.COLUMN);
		}

	for ( int i = 0; i < SudokuSolver.GRID_SIZE; i++ )
		{
		// create a set of the options block
		Cell[] block = puzzle.getBlock(i);

		f += resolveUniquePositions(block, Grouping.BLOCK);
		}

	return f;
	}

private int resolveUniquePositions(Cell[] sets, Grouping direction) throws IllegalGridException
	{
	int f = 0;
	int[] occur = Puzzle.getFrequencyOfPossibilities(sets);

	// find any numbers that only occur once
	for ( int i = 0; i < occur.length; i++ )
		{
		if ( occur[i] == 1 )
			{
			// find where it occurs and remove all other values
			for ( Cell s : sets )
				{
				if ( s.contains(i+1) )
					{
					Cell removed = new Cell(s);
					removed.remove(i+1);
					
					s.clear();
					s.add(i+1);
					
					fireResolved(hGrid, s, sets, "Zero Choice: No other cells in this group have this number", removed, direction);
					break;
					}
				}


			f++;
			}
		}

	return f;
	}

/**
 * fires the square resolved event to all listeners
 * 
 * @param the complete grid of possible values
 * @param loc the location of the resolved square 
 * @param group 
 * @param removed 
 * @param resolvedMsg 
 * @throws IllegalGridException 
 */
private void fireResolved(Cell[][] grid, Cell loc, Cell[] group, String msg, Cell removed, Grouping direction) throws IllegalGridException
	{
	if ( !hasSolverListeners() )
		return;
	
	List<Point> ref = new ArrayList<Point>();
	
	for ( Cell s : group )
	    {
	    // if this is not resolved and is not the location
	    if ( s.size() > 1 && s != loc )
	    	{
	    	ref.add(puzzle.getLocation(s));
	    	}
	    }
	
	SolutionEvent e = new ZeroChoiceEvent(this, msg, puzzle.clone(), puzzle.getLocation(loc), removed, ref, direction);
	fireSolutionEvent(e);
	}
}
